.wy-nav-content {
    max-width: 900px !important;
}

/* override table width restrictions */
.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: normal !important;
}

.wy-table-responsive {
    overflow: visible !important;
}