# Copyright 2019 Seth V. Neel, Michael J. Kearns, Aaron L. Roth, Zhiwei Steven Wu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


class ClassifierHistory(object):
    """Class ClassifierHistory storing lists of classifiers at each iteration and predictions."""
    def __init__(self):
        super(ClassifierHistory, self).__init__()
        self.classifiers = []
        self.predictions = []

    def append_classifier(self, classifier):
        self.classifiers.append(classifier)

    def get_most_recent(self):
        return self.classifiers[-1]